#include <iostream>
#include "RtAudio.h"
#include <math.h>
#include "voice.h"
#include <cstdlib>
#include <string>
#include "RtMidi.h"
using namespace std;

//sample type
#define SAMPLE double
//number of channels
#define MY_CHANNELS 1
//sample rate
#define MY_SRATE 44100
//RtAudio format
#define MY_FORMAT RTAUDIO_FLOAT64
//number of strings
#define NUMSTRINGS 16

//global variable for current string
int g_current = 0;

//global middle man buffer
SAMPLE g_middleman[20000];

//create array of sixteen voices
voice strings[NUMSTRINGS];

//-------------------------------------------------------------------------------------------------
//MIDI callback function
//-------------------------------------------------------------------------------------------------
void midi_callback(double deltatime, vector<unsigned char> * message, void *userData){

  unsigned int nBytes = message->size();

  //if a noteOn message is received pluck string
  if (nBytes > 0 && (int)message->at(0) == 144){
    
    //variable to hold frequency
    double frequency;
    int note = (int)message->at(1);

    //convert byte 1 (midi note) to frequency
    frequency = pow(2,(note-69) / 12.0) * 440;

    //set appropriate delay length buffer
    strings[g_current].setfreq(frequency);

    //pluck if velocity != 0
    if((int)message->at(2) != 0){
      double volume = (int)message->at(2) / 127.0;
      strings[g_current].fill(volume);
    
      //advance to next string and if greater than number of strings set it back to beginning
      g_current++;
      if(g_current >= NUMSTRINGS) g_current = 0;
    }
  }

}

//-------------------------------------------------------------------------------------------------
//audio callback function
//-------------------------------------------------------------------------------------------------
int callback( void * outputBuffer, void * inputBuffer, unsigned int numFrames,
            double streamTime, RtAudioStreamStatus status, void * data ){
	
  //cast input and output buffers
  SAMPLE * output = (SAMPLE *)outputBuffer;	

  // zero out the output
  for( int i = 0; i < numFrames; i++ ){
    output[i] = 0;
  }
	
  for (int i=0; i<NUMSTRINGS; i++){
		
    //synthesize each string
    strings[i].synth(g_middleman, numFrames);

    //gather them all into the output buffer
    for(int j = 0; j < numFrames; j++){

      output[j] += g_middleman[j];
    }
  }

  return 0;
}

//------------------------------------------------------------------------------------------------
//Main program
//------------------------------------------------------------------------------------------------
int main(int argc, char * argv[]){

  //display welcom message and usage
  cout << "Welcome to karpluck_wqh, a very simple karplus strong implementation with MIDI input." << endl;
  cout << "Usage ./karpluck_wqh portnumber. If no port specified, default is zero. " << endl;

  //initialize random seed
  srand (time(NULL));
	
  //Create pointer to RtMidi and RtAudio
  RtAudio * adac = NULL;
  RtMidiIn * midin = NULL;
  
  try{
    midin = new RtMidiIn();
  }
  catch(RtError & err){
    err.printMessage();
    goto cleanup;
  }

  //check available ports
  if(midin->getPortCount() == 0){
    cout << "No ports available!" << endl;
    delete midin;
    exit(1);
  }
  
  //check to see if a port has been specified
  if(argc == 2){
    //if so, open the specified port
    int port = atoi(argv[1]);
    midin->openPort(port);
  }
  else{
  //use first port available
  midin->openPort(0);
  }

  //set midi callback
  midin->setCallback(&midi_callback);

  //don't ignore warnings (LLLLLLLLLAAAAAAAMMMMMMEEEEEEEE)
  midin->ignoreTypes(false,false,false);

  //initialize RtAudio
  try{
    adac = new RtAudio();
  }
  catch(RtError & fart){
    fart.printMessage();
    goto cleanup;
  }

  //initialize the strings
  for(int i = 0; i < NUMSTRINGS; i++){
    strings[i].initialize(MY_SRATE*2, MY_SRATE, 440);
  }  

  try{
    //buffer frames and bytes
    unsigned int bufferBytes = 0;
    unsigned int bufferFrames = 512;
    
    //check to see if an audio device is connected
    if (adac->getDeviceCount() < 1){
      cout << "No audio device detected!" << endl;
      exit(1);
    }
      
    //show RtAudio warning on console
    adac->showWarnings(true);

    //set up i/o parameters
    RtAudio::StreamParameters oParams;
    oParams.deviceId = adac->getDefaultOutputDevice();
    oParams.nChannels = MY_CHANNELS;
    oParams.firstChannel = 0;
    
    //stream options
    RtAudio::StreamOptions options;
    
    //open a stream
    adac->openStream (&oParams, NULL, MY_FORMAT, MY_SRATE, &bufferFrames, &callback, NULL, &options);
    adac->startStream();
  }
  catch (RtError& e){
    //houston we have a problem
    cout << e.getMessage() << endl;
    goto cleanup;
  }

  //wait for the notes
  char input;
  cout << "Start mashin them keys! Hit <enter> to quit." << endl;
  cin.get(input);

  //stop the stream
  try{
    //stop the stream
    adac->stopStream();
  }
  catch(RtError & poop){
    poop.printMessage();
  }
    
 cleanup:
  if(adac->isStreamOpen())
    adac->closeStream();
  if(midin)
    delete midin;
  
  //finito
  return 0;
}
